###
# Pod Documentation
###

=head1 NAME

MummerDelta.pm

=head1 SYNOPSIS

Module for reading and writing MUMmer delta files.

=head1 DESCRIPTION

=head1 AUTHOR

Jeffrey Barrick
<jbarrick@msu.edu>

=head1 COPYRIGHT

Copyright 2008.  All rights reserved.

=cut

###
# End Pod Documentation
###

package MummerDelta;
use vars qw(@ISA);
use strict;

use Bio::Root::Root;
use Data::Dumper;


@ISA = qw( Bio::Root::Root );

=head2 new

 Title   : new
 Usage   : $diff = Intergenic::454Diff->new( -file_name => 'sequence.HCDiff.txt' );
 Function: Creates a diff object and initializes it from a file.
 Returns : a new 454Diff object

=cut

sub new
{
	my($caller,@args) = @_;
	my $class = ref($caller) || $caller;
	my $self = new Bio::Root::Root($caller, @args);

	bless ($self, $class);
	$self->{file_name} = $self->Bio::Root::RootI::_rearrange([qw(FILE_NAME)], @args);
	$self->{file_name} or $self->throw("Must provide -file_name parameter");
	#$self->{dont_reverse} = $self->Bio::Root::RootI::_rearrange([qw(dont_reverse)], @args);
		
	#opening for writing
	if ($self->{file_name} =~ m/^>/)
	{
		open $self->{fh}, "$self->{file_name}";
	}
	
	#opening for reading
	else
	{	
		#open file and read past two header lines
		open $self->{fh}, "$self->{file_name}";
		$self->{header_1} = readline $self->{'fh'};
		$self->{header_2} = readline $self->{fh};
		$self->{next_line} = readline $self->{fh};
	}
	
	return $self;
}

=head2 load_file

 Title   : load_file
 Usage   : Intergenic::454Diff->create_table();
 Function: Called to create a new table, override to get a more complex table
 Returns :

=cut

sub get_next
{
	my ($self) = @_;

	return undef if (!$self->{next_line}); #done reading
	$self->{fh} or $self->throw("No filehandle exists.");
	
	my $l = $self->{next_line};
	chomp $l;
	my @ll = split /\s+/, $l; 
	
	my $item;
	$item->{reference} = $ll[0];
	$item->{reference} =~ s/^>//;
	$item->{query} = $ll[1];
	$item->{reference_length} = $ll[2];
	$item->{query_length} = $ll[3];

	while ($l = readline $self->{fh})
	{
		last if ($l =~ m/^>/);
		
		chomp $l;
		@ll = split /\s+/, $l;
		my $match;
		$match->{reference_start} = $ll[0];
		$match->{reference_end} = $ll[1];
		$match->{query_start} = $ll[2];
		$match->{query_end} = $ll[3];
		$match->{mismatches} = $ll[4];
		$match->{extra_1} = $ll[5];
		$match->{extra_2} = $ll[6];
		
		#reverse to be relative to query
		$match->{strand} = +1;
		if ($match->{query_start} > $match->{query_end})
		{
			$match->{strand} = -1;
			($match->{query_start}, $match->{query_end}) = ($match->{query_end}, $match->{query_start});
		}

		@{$match->{indels}} = ();
		
		while ($l = readline $self->{fh})
		{
			chomp $l;
			last if ($l == 0); 
			push @{$match->{indels}}, $l;
		}
		
		#print STDERR Dumper($match);
		push @{$item->{matches}}, $match;
	}
	
	$self->{next_line} = $l;
	return $item;
}

sub write_item
{
	my ($self, $i) = @_;
	my $fh = $self->{fh};
	
	
	print $fh join(' ', ">$i->{reference}",$i->{query},$i->{reference_length},$i->{query_length}) . "\n";
	foreach my $m (@{$i->{matches}})
	{
		my ($start, $end) = ($m->{query_start},$m->{query_end});
		($start, $end) = ($end, $start) if ($m->{strand} == -1);
		
		print $fh join(' ', "$m->{reference_start}",
			$m->{reference_end},$start,$end,
			$m->{mismatches},$m->{extra_1},$m->{extra_2}) . "\n";
			
		foreach my $d (@{$m->{indels}})
		{
			print $fh "$d\n";
		}
		
		print $fh "0\n";
	}
}

sub write_header
{
	my ($self, $h) = @_;
	print {$self->{fh}} "$h->{header_1}$h->{header_2}";
}


#static method for determining if something is unique
sub to_unique
{
	my ($item) = @_;
	$item or die "Undefined input.";

	my $best_match;
	my $min_mismatches = $item->{query_length}+1;
	foreach my $m (@{$item->{matches}})
	{
		if ($m->{mismatches} == $min_mismatches)
		{
			undef $best_match;
		}
		elsif ($m->{mismatches} <= $min_mismatches)
		{
			$best_match = $m;
			$min_mismatches = $m->{mismatches};
		}
	}
	
	return undef if (!$best_match);
	
	@{$item->{matches}} = ();
	push @{$item->{matches}}, $best_match;
	
	return $item;
}

#find the match of the item closest to the input distance
sub to_pair
{
	my ($item, $start, $end, $distance) = @_;
	$item or die "Undefined input.";

	my $best_match;
	my $min_mismatches = $item->{query_length}+1;
	foreach my $m (@{$item->{matches}})
	{
		if ($m->{mismatches} == $min_mismatches)
		{
			undef $best_match;
		}
		elsif ($m->{mismatches} <= $min_mismatches)
		{
			$best_match = $m;
			$min_mismatches = $m->{mismatches};
		}
	}
	
	return undef if (!$best_match);
	
	@{$item->{matches}} = ();
	push @{$item->{matches}}, $best_match;
	
	return $item;
}

sub to_best
{
	my ($item) = @_;
	$item or die "Undefined input.";

	@{$item->{matches}} = sort by_length_mismatch @{$item->{matches}};
	
	sub by_length_mismatch
	{
		return (-(($a->{query_end} - $a->{query_start}) <=> ($b->{query_end} - $b->{query_start}))
		|| ($a->{mismatches} <=> $b->{mismatches}));
	}
		
	my @new_matches;
	my @matched_pos;
	
	MATCH: foreach my $m (@{$item->{matches}})
	{
		$m->{redundant} = 0;
		my $max_overlap = 0;
		
		#check for overlap with previously accepted
		
		my $accept = 1;
		TEST_MATCH: foreach my $c (@new_matches)
		{
			my $overlap = 0;

			#accept same length, same mismatches, outright, but call them redundant then
			if (($c && ( $m->{query_end} - $m->{query_start} == $c->{query_end} - $c->{query_start})
				&& ($m->{mismatches} == $c->{mismatches})))
			{
				$c->{redundant} = 1;
				$m->{redundant} = 1;
				last TEST_MATCH;
				#accepted
			}
			else
			{

#				if ( ($c->{query_start} >= $m->{query_start}) && ($c->{query_start} <= $m->{query_end}) )
#				{
#					$overlap = $m->{query_end} - $c->{query_start} + 1;
#				}
#				elsif ( ($m->{query_start} >= $c->{query_start}) && ($m->{query_start} <= $c->{query_end}) )
#				{
#					$overlap = $c->{query_end} - $m->{query_start} + 1;
#				}
#
#				
#				if ($overlap > $max_overlap)
#				{
#					$max_overlap = $overlap;
#					last TEST_MATCH if ($max_overlap > 2);
#				}
				
				
			}
		}
		
		if ($m->{redundant} == 0)
		{
			my $test_matched_pos = 0;
			for my $pos ($m->{query_start}-1..$m->{query_end}-1)
			{
				$test_matched_pos++ if ($matched_pos[$pos]);
			}
			my $test_overlap_pct = $test_matched_pos / ($m->{query_end} - $m->{query_start} + 1);
			
			if ($test_overlap_pct > 0.5)
			{
				$accept = 0;
			}
		}		
		
		if ($accept)
		{
			push @new_matches, $m;
			
			for my $pos ($m->{query_start}-1..$m->{query_end}-1)
			{
				$matched_pos[$pos]++;
			}
		}
	}
	
	#die if ($item->{query} == 45);
	$item->{query_pos_matched} = 0;
	foreach my $count (@matched_pos)
	{
		$item->{query_pos_matched}++ if ($count);
	}
	
	@{$item->{matches}} = @new_matches;
	return $item;
}


sub remove_dominated_matches
{
	my ($item) = @_;
	$item or die "Undefined input.";

	@{$item->{matches}} = sort by_length_mismatch @{$item->{matches}};

	my @new_matches;
	MATCH: foreach my $m (@{$item->{matches}})
	{			
		##check for being dominated by other matches
		TEST_MATCH: foreach my $c (@new_matches)
		{			
			## ditch if it is contained
			if ( ($m->{query_start} >= $c->{query_start}) && ($m->{query_end} <= $c->{query_end}) )
			{
				## and shorter
				if ($m->{query_end} - $m->{query_start} < $c->{query_end} - $c->{query_start})
				{
					next MATCH;
				}
			}
		
			## ditch if it has the same start and end but more mismatches
			if ( ($m->{query_start} >= $c->{query_start}) && ($m->{query_end} <= $c->{query_end}) )
			{				
				next MATCH if ($m->{mismatches} > $c->{mismatches});
				next MATCH if ( ($m->{query_start} > $c->{query_start}) || ($m->{query_end} < $c->{query_end}) );
			}
		}
		## accept it
		push @new_matches, $m;
	}
	
	@{$item->{matches}} = @new_matches;
	return @new_matches;
}

sub to_best_with_redundancy
{

# The objective is to find the minimum number of separate categories of
# matches that cover as much of the read as possible.

# When the read matches one position in the reference genome, this is straightforward.
# The trick is in hybrid reads that match multiple, separate locations in the genome.
# Watch out for cases where a read mostly overlaps the already accepted ones, and only adds a few nt.
#   In this case, favor the one that is closer to an end of the read.
# We must also watch out for cases where an entire category of match is redundanct (IS elements).
# Later, the overlap between these separate matches must be resolved in a principled manner!

# sort in descending order by length
# add longest, keeping multiple if 

	my ($item) = @_;
	$item or die "Undefined input.";

	@{$item->{matches}} = sort by_length_mismatch @{$item->{matches}};
	my @best_lists;
	
	MATCH: foreach my $m (@{$item->{matches}})
	{		
		#check for overlap with previously accepted
		LIST: foreach my $list (@best_lists)
		{
			TEST_MATCH: foreach my $c (@$list)
			{			
				my $overlap_percent = 0;
				if ( ($m->{query_start} >= $c->{query_start}) && ($m->{query_start} <= $c->{query_end}) )
				{
					$overlap_percent = ($c->{query_end} - $m->{query_start}) / ($m->{query_end} - $m->{query_start});
				}
				elsif ( ($m->{query_end}   >= $c->{query_start}) && ($m->{query_end}   <= $c->{query_end}) )
				{
					$overlap_percent = ($m->{query_end} - $c->{query_start}) / ($m->{query_end} - $m->{query_start});
				}

				#what about same length matches?
				#Save same length, same mismatch matches
				if ($overlap_percent > 0.5)
				{
					if (($c->{query_end} - $c->{query_start} == $m->{query_end} - $m->{query_start}) && ($m->{mismatches} == $c->{mismatches})) 
					{
						push @$list, $m;
					}
					next MATCH;
				}	
			}
		}
		
		#new list
		my @new_list;
		push @new_list, $m;
		push @best_lists, \@new_list;
	}
	
	#now_convert back to a normal list, but classify redundancy of hit...
	my @new_matches;
	foreach my $list (@best_lists)
	{
		foreach my $match (@$list)
		{
			$match->{redundancy} = scalar @$list;
			push @new_matches, $match;
		}
	}
	
	@{$item->{matches}} = @new_matches;
	return @best_lists;
}


sub to_complete_query_matches
{
	my ($item) = @_;
	$item or die "Undefined input.";

	#print Dumper($item);

	##allow no indels
	@{$item->{matches}} = grep { scalar @{$_->{indels}} == 0 } @{$item->{matches}};
	return $item if (scalar @{$item->{matches}} == 0);
	
	##keep those that completely cover the query
	##and have the minimum number of mismatches
	@{$item->{matches}} = sort by_length_mismatch @{$item->{matches}};
	
	my $min_mismatches = $item->{matches}->[0]->{mismatches};
	@{$item->{matches}} = grep { ($_->{mismatches} == $min_mismatches) and ($_->{query_end} - $_->{query_start} + 1 == $item->{query_length}) } @{$item->{matches}};
	
	return $item;
}


sub get_best_lists
{

# sort in descending order by length
# add longest, keeping multiple if 

	my ($item) = @_;
	$item or die "Undefined input.";

	@{$item->{matches}} = sort by_length_mismatch @{$item->{matches}};
		
	my @best_lists;
	
	MATCH: foreach my $m (@{$item->{matches}})
	{		
		#check for overlap with previously accepted
		LIST: foreach my $list (@best_lists)
		{
			TEST_MATCH: foreach my $c (@$list)
			{
				my $overlap_percent = 0;
				if ( ($m->{query_start} >= $c->{query_start}) && ($m->{query_start} <= $c->{query_end}) )
				{
					$overlap_percent = ($c->{query_end} - $m->{query_start}) / ($m->{query_end} - $m->{query_start});
				}
				elsif ( ($m->{query_end}   >= $c->{query_start}) && ($m->{query_end}   <= $c->{query_end}) )
				{
					$overlap_percent = ($m->{query_end} - $c->{query_start}) / ($m->{query_end} - $m->{query_start});
				}

				#what about same length matches?
				#Save same length matches
				if ($overlap_percent > 0.4)
				{
					push @$list, $m if ($c->{query_end} - $c->{query_start} == $m->{query_end} - $m->{query_start}); 
					next MATCH;
				}	
			}
		}
		
		#new list
		my @new_list;
		push @new_list, $m;
		push @best_lists, \@new_list;
	}
	return @best_lists;
}

sub filter_homopolymeric
{
	## read // reference to sequence which has been split into a list // length below which to do filtering
	my ($match_list_ref, $s, $length_cutoff) = @_;
	$length_cutoff = 22 if (!defined $length_cutoff);
	
	my @filtered_matches;
	MATCH: foreach my $m (@$match_list_ref)
	{
		##don't examine matches that are above a certain length threshold to save time 
		next MATCH if ($m->{reference_end} - $m->{reference_start} + 1 > 22);

		#my $subseq = $seq->subseq($m->{reference_start}, $m->{reference_end});
		my @subseq = @{$s}[$m->{reference_start}-1..$m->{reference_end}-1];
		my %nt_hash;
		foreach my $nt (@subseq)
		{
			$nt_hash{$nt}++;
		}
		my @keys = sort { $nt_hash{$a} <=> $nt_hash{$b} } keys %nt_hash;
		#my $len = length $subseq;
		my $len = scalar @subseq;
		my $top = $nt_hash{$keys[-1]};
		
		my $diff_nt = scalar @keys;
		if ($diff_nt <= 2)
		{
			#print STDERR "Only two nucleotides: $m->{reference_start}\-$m->{reference_end}  $diff_nt\n";
			#print STDERR "@subseq\n";
			next MATCH;				
		}
		
		if ($top / $len > 0.8) 
		{
			#print STDERR "Homopolymer: $m->{reference_start}\-$m->{reference_end}  $top / $len\n";
			#print STDERR "@subseq\n";
			next MATCH;
		}	
		
		push @filtered_matches, $m;
	}
	
	return @filtered_matches;
}

#sub get_best_lists
#{
#	my ($item) = @_;
#	$item or die "Undefined input.";
#
#	@{$item->{matches}} = sort by_length_mismatch @{$item->{matches}};
#		
#	my @best_lists;
#	my @new_matches;
#	
#	MATCH: foreach my $m (@{$item->{matches}})
#	{
#		$m->{redundant} = 0;
#		my $max_overlap = 0;
#		
#		#check for overlap with previously accepted
#		LIST: foreach my $list (@best_lists)
#		{
#			TEST_MATCH: foreach my $c (@$list)
#			{
#				my $overlap = 0;
#
#				#find out the overlap
#				if ( ($c->{query_start} >= $m->{query_start}) && ($c->{query_start} <= $m->{query_end}) )
#				{
#					$overlap = $m->{query_end} - $c->{query_start} + 1;
#				}
#				elsif ( ($m->{query_start} >= $c->{query_start}) && ($m->{query_start} <= $c->{query_end}) )
#				{
#					$overlap = $c->{query_end} - $m->{query_start} + 1;
#				}
#
#				#if there was sufficient overlap, and it does not have more mismatches
#				#than already added hits in this list, then save it.
#				if ($overlap > 2)
#				{
#					push @$list, $m if ($m->{mismatches} <= $c->{mismatches});
#					next MATCH;
#				}
#			}
#		}
#		
#		#new list
#		my @new_list;
#		push @new_list, $m;
#		push @best_lists, \@new_list;
#	}
#	return @best_lists;
#}


return 1;
