#!/usr/bin/perl -w

###
# Pod Documentation
###

=head1 NAME

breseq.pl

=head1 SYNOPSIS

Usage: breseq.pl [-a] [-q solexa|phred] reference.gbk reads.fastq

Perform an analysis of bacterial resequencing data.

=head1 DESCRIPTION

=over

=item B<-a>

Print additional intermediate files.

=item B<-q>

Style of quality scores in fastq file. Must be either "solexa" (default) or "phred".
Generally, if the largest ASCII values you see are no greater than 70 (=F) then you have "phred" style scores.
If you have "solexa" style scores, then you probably see lowercase letters (=a). 

=back

=head1 AUTHOR

Jeffrey Barrick
<jbarrick@msu.edu>

=head1 COPYRIGHT

Copyright 2008.  All rights reserved.

=cut

###
# End Pod Documentation
###

## TODO
## Rewrite to ignore N's?
## Rewrite scoring scheme to give final scores resembling Solexa and more proper calculations of evidence for/against.
## Add auto-detection of score style.

use strict;
use Data::Dumper;
#print Dumper(%ENV);

use FindBin;
use lib $FindBin::Bin;
$ENV{PATH} = "$ENV{PATH}:" . $FindBin::Bin;

use MummerDelta;
use Bio::SeqIO;

our $VERSION = '0.00_01';

use POSIX qw(ceil floor);

our %base_to_int = ( 'A'=>0, 'C'=>1, 'T'=>2, 'G'=>3, '.'=>4, 'N'=>5 );
our @int_to_base = ('A','C','T','G','.','N');
our $quality_cutoff = 10;
our $quality_type = 'solexa';
our $ignore_short_match_cutoff = 16;

#Get options
use Getopt::Long;
use Pod::Usage;
my ($help, $man);
my ($verbose);
my $write_all_files;
GetOptions(
	'help|?' => \$help, 'man' => \$man,
	'verbose|v' => \$verbose,
	'write-all-files|a' => \$write_all_files,
	'quality-style|q=s' => \$quality_type,
	'quality-cutoff|c=s' => \$quality_cutoff,
) or pod2usage(2);
pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

die "Unrecognized quality score style: $quality_type" if (($quality_type ne 'solexa') && ($quality_type ne 'phred'));
our $using_solexa_quality = ($quality_type eq 'solexa');
print STDERR "Using fastq quality score style: $quality_type\n";
print STDERR "SNP quality cutoff: $quality_cutoff\n";

my ( $ref_genbank, $read_fastq ) = @ARGV;

my $command;
my $res;

## Convert the read fastq file to fasta for input into 
my $read_fasta = $read_fastq;
$read_fasta =~ s/\.fastq$//;
$read_fasta .= ".fasta";
if (!-e $read_fasta)
{
	print STDERR "Converting read fastq to fasta\n";
	$command = "fastq_to_fasta.pl $read_fastq $read_fasta";
	#print "$command\n";
	$res = system $command;
	die "$res" if ($res);
}

##
# Load the Input Genbank file
##
print STDERR "Loading reference sequence(s) from file: $ref_genbank\n";
my $ref_i = Bio::SeqIO->new( -file => "<$ref_genbank");

my $ref_fasta = "reference.fasta";
my $create_ref_fasta = ((!-e $ref_fasta) || (-s $ref_fasta == 0));

my $ref_o;
$ref_o = Bio::SeqIO->new( -file => ">$ref_fasta", -format => 'fasta') if ($create_ref_fasta);

my %ref_seqs;
my %ref_strings;
my %gene_lists;
my %is_lists;

### We will need to eventually allow multiple genomic fragments
### in the genbank file in the rest of the code.

while (my $ref_seq = $ref_i->next_seq)
{
	my $id = $ref_seq->id;
	print STDERR "  Processing $id:\n";

	$ref_seqs{$id} = $ref_seq;
	
	if ($create_ref_fasta)
	{
		print STDERR "    Creating fasta file...\n";
		$ref_o->write_seq($ref_seq);
	}
	else
	{
		print STDERR "    Fasta file already exists...\n";
	}
	
	print STDERR "    Loading genes...\n";

	##it is much faster to use substr to create the lists for nt comparisons than BioPerl trunc
	$ref_strings{$id} = $ref_seq->seq;
	$ref_strings{$id} = "\U$ref_strings{$id}"; ##uppercase for comparisons

	#load the genbank file
	my @Feature_List = $ref_seq->get_SeqFeatures();
	my @gene_list;
	my @is_list;
	
	FEATURE: foreach my $Feature ( @Feature_List ) 
	{ 	
		if ($Feature->primary_tag eq 'repeat_region')
		{	
			my $is;
			$is->{gene} = GetTag($Feature, "mobile_element");		
			$is->{start} = $Feature->start;
			$is->{end} = $Feature->end;
			$is->{strand} = $Feature->strand;
			push @is_list, $is;
			next FEATURE;
		}
	
#		if ($Feature->primary_tag eq 'gene')
#		{	
#			my $gene;
#			$gene->{gene} = GetTag($Feature, "gene");
#			
#			$gene->{start} = $Feature->start;
#			$gene->{end} = $Feature->end;
#			$gene->{strand} = $Feature->strand;
#			$gene->{product} = "";
#			push @gene_list, $gene;
#		}
		
		## add additional information to the last 
		if (   ($Feature->primary_tag eq 'CDS') 
			|| ($Feature->primary_tag eq 'tRNA') 
			|| ($Feature->primary_tag eq 'rRNA') )

		{
			#Add information to last gene
			
			my $gene;
			$gene->{gene} = GetTag($Feature, "gene");
			$gene->{gene} = GetTag($Feature, "locus_tag") if (!$gene->{gene});
			$gene->{start} = $Feature->start;
			$gene->{end} = $Feature->end;
			$gene->{strand} = $Feature->strand;
			$gene->{product} = "";
			$gene->{note} = GetTag($Feature, "note");
				
			$gene->{accession} = GetTag($Feature, "protein_id");
			$gene->{translation} = GetTag($Feature, "translation");
			$gene->{product} = GetTag($Feature, "product");
			
			#set a type for the feature
			$gene->{type} = $Feature->primary_tag;
			$gene->{type} = "protein" if ($gene->{type} eq 'CDS');
			$gene->{pseudogene} = GetTag($Feature, "pseudo");
			$gene->{type} = "pseudogene" if ($gene->{pseudogene});
			
			##assume if there is no translation that we have a pseudogene...
			$gene->{type} = "pseudogene" if (($Feature->primary_tag eq 'CDS') && (!$gene->{translation}));
			$gene->{index} = scalar @gene_list;
			
			push @gene_list, $gene;		
		}
	}
	
	$gene_lists{$id} = \@gene_list;
	$is_lists{$id} = \@is_list;
}

# Calculate nucmer parameters
my $r = 36; 	#min read length
my $d = 1;		#max differences

##parameters for forcing the entire read to match one position
#my $minmatch = ceil(($r-$d) / ($d+1));
#my $mincluster = $minmatch;
#my $maxgap = $d;
#my $breaklen = $r - $minmatch;

##minimum size of exact match...
my $minmatch = 14;

##minimum size of match to be extended...
my $mincluster = $minmatch;

#breaklength is how far we force extension to the end of the read
#it is only used if --nooptimize is enabled
my $breaklen = 0;


##Run mummer
	
##Determine deletions with strict matching of reads.
#if (! -e "mummer.exact.delta")
#{
#	$command = "nucmer --maxmatch -l 14 -c 14 -g 0 --noextend --prefix mummer.exact $ref_fasta $read_fasta";
#	print "$command\n";
#	$res = system $command;
#	die "$res" if ($res);
#}
	
##Determine SNPs and rearrangements from looser matching
if (! -e "mummer.delta")
{
	print STDERR "Performing MUMmer alignment...\n";
	$command = "nucmer --maxmatch -c $mincluster  -l $minmatch --prefix mummer $ref_fasta $read_fasta";	 ## -b $breaklen --nooptimize
	print STDERR "$command\n";
	$res = system $command;
	die "$res" if ($res);
	
	##we don't need the cluster file, remove to save space
	unlink "mummer.cluster";
}
else
{
	print STDERR "MUMmer output already exists...\n";
}
	
	
## Information that we need to 
## save about each SNP
##
## probability and coverage by strand of each base
## probability and coverage by strand of linkage between this nt and the next one	
## unique and redundant coverage of this position by
	
##
# Correct reads, record quality scores and coverage per position
##

##placeholder, will be used to handle multiple reference fragments at some point
my @fragment_ids = sort keys %ref_strings;
my $seq_id = $fragment_ids[0];

## strategy, we create a file that gives coverage information for the reference sequence
## this has a fixed length, we don't have to store it in memory.
## we create multiple files for each SNP that DOES NOT AGREE with this file
## write them all all out, sort them in reverse, and then read them backwards, optionally truncating as we go
my $snp_diff_file= 'snp_diff.dat';

our $cov;
my $coverage_file = 'coverage.dat';
our $coverage_item_format = "ffffSSSSff";
my $test_size = pack $coverage_item_format, (0,0,0,0,0,0,0,0,0,0);
our $coverage_item_size = length $test_size;
  		 
if (!-e $coverage_file)
{
	print STDERR "Correcting ambiguous alignments and comparing sequences...\n";
	
	##allocate a string big enough to hold data
	$cov = "\0" x (length($ref_strings{$seq_id}) * $coverage_item_size);
	
	#open the diff file
	my $md = MummerDelta->new(-file_name => "mummer.delta", -dont_reverse => 1);
	my $read = $md->get_next;

	#open the read fastq file
	my $in = Bio::SeqIO->new( -file => "<$read_fastq", -format => "fastq", -verbose => -1);
	my $read_seq = $in->next_seq;

	#open the corrected delta file (optional)
	my $mdo;
	$mdo = MummerDelta->new(-file_name => ">mummer.delta.corrected") if ($write_all_files);

	#open the split delta file (optional)
	my $mds;
	$mds = MummerDelta->new(-file_name => ">mummer.delta.split") if ($write_all_files);

	my @snps;
	
	my $done_reads = 0; 
	
	READ: while (my $read = $md->get_next)
	{	
		## Keep a count, so we can print out progress...
		$done_reads++;
		print STDERR "  $done_reads " . (scalar @snps) . "\n" if ($done_reads % 1000 == 0);
		#last if ($done_reads % 10000 == 0);
	
		#We are only concerned with "non-dominated" matches.
		MummerDelta::remove_dominated_matches($read);		

		#move forward until we reach the correct read in the fastq file
		while ($read_seq->id ne $read->{query})
		{
			$read_seq = $in->next_seq;
			die "Could not find sequence $read->{query}" if (!$read_seq);
		}
		$read_seq->force_flush(1); #necessary for reading fastq files that do not label qualities?

		##Loop through the matches and determine which have matched the exact same sequence
		##so they can be treated as a group.
		my @match_groups;
		MATCH: foreach my $m (@{$read->{matches}})
		{
			my $ref_seq_matched = substr $ref_strings{$seq_id}, $m->{reference_start}-1, $m->{reference_end}-$m->{reference_start}+1;
		
			#my $ref_seq_matched = ($m->{strand} == +1)
			#	? substr $ref_strings{$seq_id}, $m->{reference_start}-1, $m->{reference_end}-$m->{reference_start}+1;
			#	: substr $ref_strings_revcom{$seq_id}, $m->{reference_start}-1, $m->{reference_end}-$m->{reference_start}+1;

			#get the sequence of the read
			GROUP: foreach my $g (@match_groups)
			{
				#start and end coords must be the same
				if ( ($m->{query_start} == $g->{matches}->[0]->{query_start}) && ($m->{query_end} == $g->{matches}->[0]->{query_end}) )
				{
					#sequence matched in the reference genome must be the same
					if ($ref_seq_matched eq $g->{ref_seq_matched}->{$m->{strand}})
					{
						push @{$g->{matches}}, $m;
						next MATCH;
					}
				}
			}
			
			##if we got here, we need a new group
			
			my $ng;
			$ng->{ref_seq_matched}->{$m->{strand}} = $ref_seq_matched;
			
			my $tmp_seq = Bio::Seq->new(-seq => $ref_seq_matched);
			$ng->{ref_seq_matched}->{-$m->{strand}} = $tmp_seq->revcom()->seq;	
			undef $tmp_seq;
			
			#create the read sequence on the top strand
			#must create a new because reversing the meta data of fastq sequence doesn't work properly...
		#	$tmp_seq = Bio::Seq->new(-seq => $read_seq->seq);
		#	$tmp_seq = $tmp_seq->trunc($m->{query_start}, $m->{query_end});
			
		#	$ng->{qry_seq_matched}->{$m->{strand}} = $tmp_seq->seq;
		#	$ng->{qry_seq_matched}->{-$m->{strand}} = $tmp_seq->revcom()->seq;		
			
			push @{$ng->{matches}}, $m;
			push @match_groups, $ng;
		}
						
		#Now is our chance to decide which groups of matches are compatible,
		#to change their boundaries and to come up with a final list.
		if (scalar @match_groups > 1)
		{
			#print STDERR "#####\n";
			#print STDERR Dumper(@match_groups);
			
			MATCH_GROUP: foreach my $mg (@match_groups)
			{
			}
			
			##### FOR NOW IGNORE THEM!!!! #####
			next READ;
		
		
			## Generally, require two groups to span most of the read
			##
			
		
			## Require at least one side to be unique in the reference genome.
			## this may ignore some junctions, but helps eliminate
			
			
			# we need to do some filtering for homopolymeric reads
			## need to check how this could mess up normal coverage of the E. coli genome
		}
		
		
		##Ignore reads with a single homopolymeric match
		##These are often homopolymeric reads.
		if (scalar @match_groups == 1)
		{
			@match_groups = filter_homopolymeric(\@match_groups);
		}
		
		## We need to rule out short matches, they are often spurious
#		my %read_coverage;
#		MATCH_GROUP: foreach my $mg (@match_groups)
#		{			
#			MATCH: for (my $i=0; $i< scalar @{$mg->{matches}}; $i++)
#			{
#				my $m = $mg->{matches}->[$i];
#				for my $i ($m->{query_start}..$m->{query_end})
#				{
#					$read_coverage{$i}++;
#				}
#			}
#		}
#		
#		if (scalar keys %read_coverage <= 0.75 * $read->{query_length}) 
#		{
#			print STDERR Dumper($read);
#			print STDERR $read_seq->seq() . "\n";
#			next READ;
#		}
		

		#correct matches where there is a deletion or insertion in a homopolymer
		#to shift this change to the rightmost position it could occupy (relative to the top strand of the genome).

		MATCH_GROUP: foreach my $mg (@match_groups)
		{			
			MATCH: for (my $i=0; $i< scalar @{$mg->{matches}}; $i++)
			{
				my $m = $mg->{matches}->[$i];
				$m->{redundancy} = scalar @{$mg->{matches}};
							
				##Keep the quality in the correct orientation
				my @quals = @{$read_seq->qual};
				@quals = reverse @quals if ($m->{strand} == -1);	
							
				##adjust the position of each indel
				##note that we have a problem detecting indels in a situation where
				##the initial or final bases occur in a homopolymeric stretch.
				##this has to be resolved at a later stage of analysis.
				
				#get the reference sequence 
				my @ref = split //, (substr $ref_strings{$seq_id}, $m->{reference_start}-1, $m->{reference_end}-$m->{reference_start}+1);
					
				#create the read sequence on the top strand
				#must create a new because reversing the meta data of fastq sequence doesn't work properly...
				my $tmp_seq = Bio::Seq->new(-seq => $read_seq->seq);
				my $qry_string = ($m->{strand} == +1) ?
					$tmp_seq->trunc($m->{query_start}, $m->{query_end})->seq :
					$tmp_seq->trunc($m->{query_start}, $m->{query_end})->revcom()->seq;
				my @qry = split //, "\U$qry_string";
				undef $tmp_seq;

					
				##
				# Shift ambiguous indels to the lowest coordinate on the top strand that is possible.
				##
				
				if (scalar @{$m->{indels}} > 0)
				{
					
					# Add indels to create the current alignment
					my $change_pos = 0;
					foreach my $indel (@{$m->{indels}})
					{
						#insert a period in the reference (if negative) or query (if positive)
						if ($indel > 0)
						{
							$change_pos += $indel;
							splice @qry, $change_pos-1, 0, '.',
						}
						else
						{
							$change_pos -= $indel;
							splice @ref, $change_pos-1, 0, '.',
						}	
					}
					print STDERR "match:\n@ref\n@qry\n" if ($verbose);

					# Find each time there is a period in a sequence and see if the query can be shifted
					my $indel_count = 0;
					my @new_indels = (0);
					my $inc = 1;
								
					# Need to look several earlier bases as long as the base is the same
					for (my $p=0; $p<scalar @ref; $p++)
					{
						if ($ref[$p] eq '.')
						{
							# If the next reference base matches the current query base, then swap the period and base
							my $found = 0;
							my $tp = $p;
							while (($tp+1 < scalar @ref) && ($ref[$tp] eq '.')) 
							{
								if ($qry[$p] eq $ref[$tp+1])
								{
									($ref[$p], $ref[$tp+1]) = ($ref[$tp+1], $ref[$p]);
									$found = 1;
									
									#$stop = 1 if ($tp > $p + 1);
									last;
								}
								$tp++;
							}				
							if (!$found)
							{
								push @new_indels, -$inc;
								$inc = 0;
							}
						}
						elsif ($qry[$p] eq '.')
						{
							# If the next qry base matches the current reference base, then swap the period and base
							
							my $found = 0;
							my $tp = $p;
							while (($tp+1 < scalar @qry) && ($qry[$tp] eq '.')) 
							{
								if ($ref[$p] eq $qry[$tp+1])
								{
									($qry[$p], $qry[$tp+1]) = ($qry[$tp+1], $qry[$p]);
									$found = 1;
									last;
								}
								$tp++;
							}				
							if (!$found)
							{
								push @new_indels, $inc;
								$inc = 0;
							}
						} 
						$inc++;
					}
					##clean up, remove the first zero from the indels
					shift @new_indels;
					$read->{matches}->[$i]->{indels} = \@new_indels;
					$m = $read->{matches}->[$i];

					print STDERR "match: strand $m->{strand}\n@ref\n@qry\n"  if ($verbose);
					print STDERR Dumper($m) if ($verbose);
				}
				
				##
				# Determine the extent of the sequence we will count as coverage
				#
				# There is also a problem detecting indels reliably in a situation where
				# the initial or final bases of a read occur in a homopolymeric stretch.
				##
				
				#correct start of query
				my $start_pos = 0;
				if ($m->{reference_start}-2 >= 1)
				{
					my $before_base = substr $ref_strings{$seq_id}, $m->{reference_start}-2, 1; 
					while ( ($start_pos < scalar @qry) && ($qry[$start_pos] eq $before_base) )
					{
						$start_pos++;
					}
				}
				
				my $after_base;  
				$after_base = substr $ref_strings{$seq_id}, $m->{reference_end}, 1 if ($m->{reference_end} < length $ref_strings{$seq_id});
				
				#correct end of query
				my $end_pos = scalar(@qry) - 1;
				if ($m->{reference_end} < length $ref_strings{$seq_id})
				{
					my $after_base = substr $ref_strings{$seq_id}, $m->{reference_end}, 1; 
					while ( ($end_pos >= 0) && ($qry[$end_pos] eq $after_base) )
					{
						$end_pos--;
					}
				}
			
				##
				# Record coverage and SNPS
				##

				## We have not passed any gaps so we can just offset past ambiguous ends
				my $qual_pos = $start_pos;
				my $ref_pos = $m->{reference_start} + $start_pos; #indexed to 1
				my $read_pos = $start_pos;
				my $insert_count = 0;
				my $paired_log10_pr = 0; #start this way so the fits nt not counted as a junction
				#pairs indicate that both N and N+1 had the same base and no insertions between
				
				#print "@ref\n@qry\n";
				## Loop through comparing each character
				
				while ($read_pos <= $end_pos)
				{
					#deletion in query relative to reference (insertion in reference)
					#note that insertions are always AFTER the position given

					if ($qry[$read_pos] eq '.')
					{			
						$insert_count = 0;
						$paired_log10_pr = 0;	
					 	push @snps, pack_snp($ref_pos, $insert_count, $quals[$qual_pos], $m->{strand}, $m->{redundancy}, $qry[$read_pos]);	
					}
					#insertion in query (deletion in reference)
					elsif ($ref[$read_pos] eq '.')
					{
						$insert_count++;
						$paired_log10_pr = 0;	
						push @snps, pack_snp($ref_pos, $insert_count, $quals[$qual_pos], $m->{strand}, $m->{redundancy}, $qry[$read_pos]);
					}
					#mismatch, no insertion
					elsif ($ref[$read_pos] ne $qry[$read_pos])
					{	
						$insert_count = 0;
						$paired_log10_pr = 0;
						push @snps, pack_snp($ref_pos, $insert_count, $quals[$qual_pos], $m->{strand}, $m->{redundancy}, $qry[$read_pos]);
					}
					else #match, update coverage
					{
						my @item;
						my $offset = ($m->{strand} == -1) ? 0 : 1;
						my $log10_pr = -log(quality_to_error_probability($quals[$qual_pos])) / log(10);
						#unique match
						if ($m->{redundancy} == 1)
						{
							$item[0+$offset] = $log10_pr;
							$item[2+$offset] = ($log10_pr + $paired_log10_pr)/2 if ($paired_log10_pr);
							$item[4+$offset] = 1;
							$item[6+$offset] = 1 if ($paired_log10_pr);
						}
						#redundant match
						else
						{
							$item[8+$offset] = 1 / $m->{redundancy};
						}

						add_coverage($ref_pos, \@item);
						$paired_log10_pr = $log10_pr;
					}
								
		 			#increment positions
					$ref_pos++ if ($ref[$read_pos] ne '.');
					$qual_pos++ if ($qry[$read_pos] ne '.');
					$read_pos++;
				}
				
			}
				
			$mdo->write_item($read) if ($write_all_files);
			#$mds->write_item($read) if ((scalar @separate_lists > 1) && ($write_all_files));

			}
	}

	##
	# Output raw SNP and coverage information
	##
	
	print STDERR "Writing coverage file...\n";
	open COV, ">$coverage_file";
	print COV $cov;
	close COV;
	undef $cov; #save some memory, because the sort may need a lot

	print STDERR "Sorting " . (scalar @snps) . " SNP diffs...\n";	
	@snps = sort { unpack("L",$a) <=> unpack("L",$b)  } @snps;
	
	print STDERR "Writing SNP diff file...\n";
	open SNP_DIFFS, ">$snp_diff_file";
	for (my $i=0; $i<scalar @snps; $i++)
	{
		print SNP_DIFFS $snps[$i];
	}
	close SNP_DIFFS;
}
else
{
	print STDERR "Correction of ambiguous alignments and comparison of sequences already complete.\n";
}


# TEMPORARY CODE TO RESORT SNP DIFFS
#{
#	my @temp_snps;
#	my $raw_snp;
#	open SNP_DIFFS, "<$snp_diff_file";
#	my $snp_diff_item_test = pack("LCCCS", 0, 0, 0, 0, 0);
#	my $snp_diff_item_size = length $snp_diff_item_test;
#	while (1)
#	{
#		my $read_bytes = read SNP_DIFFS, $raw_snp, $snp_diff_item_size;	
#		last if (!$read_bytes);
#		push @temp_snps, $raw_snp;
#
#		my $s;
#		($s->{position}, $s->{insert_count}, $s->{quality}, $s->{strand}, $s->{redundancy}, $s->{base}) = unpack_snp($raw_snp);		
#	}
#	
#	@temp_snps = sort { unpack("L",$a) <=> unpack("L",$b)  } @temp_snps;
#
#	
#	#foreach my $snp (@temp_snps)
#	#{
#	#	my $s;
#	#	($s->{position}, $s->{insert_count}, $s->{quality}, $s->{strand}, $s->{redundancy}, $s->{base}) = unpack_snp($snp);		
#	#	print Dumper($s);
#	#}	
#	close SNP_DIFFS;
#	
#	print STDERR "Writing SNP diff file...\n";
#	open SNP_DIFFS, ">$snp_diff_file";
#	for (my $i=0; $i<scalar @temp_snps; $i++)
#	{
#		print SNP_DIFFS $temp_snps[$i];
#	}
#	close SNP_DIFFS;
#}

print STDERR "Collecting and evaluating SNPs...\n";

## open extra output file 
open SNP_ALL, ">snps.all.tab" if ($write_all_files);

## read the entire coverage file, we're going to need it for coverage calculations
my $sequence_length = $ref_seqs{$seq_id}->length;
open COV, "<$coverage_file";
read COV, $cov, -s $coverage_file;
close COV;


my $snp_diff;
open SNP_DIFFS, "<$snp_diff_file";
my $snp_diff_item_test = pack("LCCCS", 0, 0, 0, 0, 0);
our $snp_diff_item_size = length $snp_diff_item_test;
read SNP_DIFFS, $snp_diff, -s $coverage_file;
close SNP_DIFFS;

my $on_snp_diff_index = 0;
my $next_snp_diff_ref;
##load the first SNP
my $num_snp_diffs = length($snp_diff) / $snp_diff_item_size;
if ($on_snp_diff_index < $num_snp_diffs)
{
	my (%s, $raw_snp);
	($s{position}, $s{insert_count}, $s{quality}, $s{strand}, $s{redundancy}, $s{base}) = get_snp_diff(\$snp_diff, $on_snp_diff_index);
	$next_snp_diff_ref = \%s;
	$on_snp_diff_index++;
	#print STDERR Dumper($on_snp_diff_index, \%s);
}

print " Sequence Length: $sequence_length\n";
print " Number of SNP Differences: $num_snp_diffs\n";

## snps is the list of items that is passed with additional information for annotating
my @snps;
my @gene_list = @{$gene_lists{$seq_id}};
my %pos_info;

## move forward until reaching a new position
my $position = 1;
my @current_raw_snps;
while ($position <= $sequence_length)
{
	print STDERR " $position\n" if ($position % 10000 == 0);

	#read the next position
	my @coverage_item = get_coverage(\$cov, $position);
	
	
	## item has 1) log10_pr -/+ strand
	##          2) log_10_pr of pair -/+ strand
	##          3) unique coverage -/+ strand
	##          4) redundant coverage -/+ strand
	
	#collect all snp diffs corresponding to this position 
	my @current_position_raw_snps;
	SNP_DIFF: while ( ($on_snp_diff_index < $num_snp_diffs) && ($next_snp_diff_ref->{position} == $position))
	{
		#save the next snp diff
		push @{$current_position_raw_snps[$next_snp_diff_ref->{insert_count}]}, $next_snp_diff_ref;
	
		#load the next one in line
		my (%s, $raw_snp);
		($s{position}, $s{insert_count}, $s{quality}, $s{strand}, $s{redundancy}, $s{base}) = get_snp_diff(\$snp_diff, $on_snp_diff_index);
		$next_snp_diff_ref = \%s;
		$on_snp_diff_index++;
		#print STDERR Dumper($on_snp_diff_index, \%s);
	}
	
	##print out information about this position and each insert count for which there were snp diffs
	
	my $last_insert_count_was_accepted = 1;
	INSERT: for (my $insert_count=0; ($insert_count < 1) or ($insert_count < scalar @current_position_raw_snps); $insert_count++)
	{	
		# complete the initial part of the line
		my $ref_base = ($insert_count == 0) ? substr($ref_strings{$seq_id}, $position-1, 1) : '.';
		my $coverage_line_part = join("\t", $coverage_item[4], $coverage_item[5], 
			sprintf("%.1f", $coverage_item[8]),sprintf("%.1f", $coverage_item[9]),
			$coverage_item[6], $coverage_item[7]);
		my $line = "$position\t$insert_count\t$ref_base\t$coverage_line_part";
		
		# zero out the info about this position
		my $pos_info;
		foreach my $base (@int_to_base)
		{
			$pos_info->{$base}->{log10_pr} = 0;
			$pos_info->{$base}->{cov}->{1} = 0;
			$pos_info->{$base}->{cov}->{-1} = 0;
		}		
		
		#add reference coverage (unique only)
		if ($insert_count == 0)
		{
			$pos_info->{$ref_base}->{log10_pr} += $coverage_item[0] + $coverage_item[1];
			$pos_info->{$ref_base}->{cov}->{-1} += $coverage_item[4];
			$pos_info->{$ref_base}->{cov}->{1}  += $coverage_item[5];
		}		
		
		#add snp difference
		foreach my $s (@{$current_position_raw_snps[$insert_count]})
		{
			next if ($s->{redundancy} > 1);
			$pos_info->{$s->{base}}->{log10_pr} += -log(quality_to_error_probability($s->{quality} )) / log(10);
			$pos_info->{$s->{base}}->{cov}->{$s->{strand}}++;
		}
		
		#print out complete information
		my $total_log10_pr = 0;
		my $best_base_log10_pr = 0;
		my $best_base = '';
		my ($total_cov, $best_cov);
		$total_cov->{1} = 0;
		$total_cov->{-1} = 0;
		$best_cov->{1} = 0;
		$best_cov->{-1} = 0;
		foreach my $base (@int_to_base)
		{			
			my $current_base_info = $pos_info->{$base};

			if ($current_base_info->{log10_pr} > $best_base_log10_pr)
			{
				$best_base = $base;
				$best_base_log10_pr = $current_base_info->{log10_pr};
				($best_cov->{-1},$best_cov->{1}) = ($current_base_info->{cov}->{-1},$current_base_info->{cov}->{1});
			}
			$total_log10_pr += $pos_info->{$base}->{log10_pr};
			
			my $log_pr = $current_base_info->{log10_pr};
			my $top_cov = $current_base_info->{cov}->{1};
			my $bot_cov = $current_base_info->{cov}->{-1};
			my $log_10_pr_string = sprintf "%.1f", $log_pr;
			$total_cov->{1}+=$top_cov;
			$total_cov->{-1}+=$bot_cov;
			
			$line .= "\t$base\t" . sprintf("%.1f",$log_10_pr_string) . "\t($bot_cov/$top_cov)";
		}
		print SNP_ALL "$line\n" if ($write_all_files);
		
		## evaluate whether to call an actual snp!
				
		my $snp;
		$snp->{start} = $position;
		$snp->{end} = $position;
		$snp->{ref_seq} = $ref_base;
		$snp->{new_seq} = $best_base;
		
		$snp->{quality} = ($insert_count > 0) 
			? $best_base_log10_pr - ($coverage_item[2] + $coverage_item[3])
			: $best_base_log10_pr - ($total_log10_pr - $best_base_log10_pr);
		
		#for insertions relative to the reference, add the pair coverage
		if ($insert_count > 0) 
		{
			 $total_cov->{-1} += $coverage_item[6];
			 $total_cov->{1} += $coverage_item[7];
		}
		$snp->{total_coverage_string} = $total_cov->{-1} . "/" . $total_cov->{1};
		$snp->{best_coverage_string} = $best_cov->{-1} . "/" . $best_cov->{1};

		$snp->{insert_index} = $insert_count;
		$snp->{info} = $line;
		$snp->{id} = $seq_id;
		
		#$snp->{coverage_string} = $paired_bases->{-1}->[$pos] . "/" . $paired_bases->{1}->[$pos];

		if ($last_insert_count_was_accepted && ($snp->{ref_seq} ne $snp->{new_seq}) && ($snp->{quality} > $quality_cutoff))
		{
			$last_insert_count_was_accepted = 1;
			print Dumper($snp);
			push @snps, $snp;
		}
		else
		{
			$last_insert_count_was_accepted = 0;
		}
		
		#we always need to allow an insert count of one
		$last_insert_count_was_accepted = 1 if ($insert_count == 0);
	}
	
	
} continue {
	$position++;
}


##
# Find deletions
##

print STDERR "Searching for deletions...\n";
print STDERR "  Estimating average coverage...\n";

my $num_positions_with_nonzero_coverage = 0;
my $average_coverage;

for (my $i=0; $i<$sequence_length; $i++)
{
	my @coverage_item = get_coverage(\$cov, $i);
	
	next if ( ($coverage_item[4] + $coverage_item[5] == 0) && ($coverage_item[8] + $coverage_item[9] == 0) );
	
	$num_positions_with_nonzero_coverage++;
	$average_coverage+=$coverage_item[4]+$coverage_item[5]+$coverage_item[8]+$coverage_item[9];
}
$average_coverage/=$num_positions_with_nonzero_coverage;
my $deletion_propagation_cutoff = $average_coverage / 5;
my $deletion_seed_cutoff = 0;

print STDERR "  Average coverage: $average_coverage\n";
print STDERR "  Deletion seed cutoff: $deletion_seed_cutoff\n";
print STDERR "  Deletion propagation cutoff: $deletion_propagation_cutoff\n";

#my $deletion_file_name = "deletions.tab";
#open DEL, ">$deletion_file_name";
#print DEL "#missing coverage\tstart\tend\tsize\tleft\tleft_inside\tright_inside\tright\n";

my $last_deletion_start_position;
my $last_coverage_item;

my $this_deletion_reaches_seed_value = 0;
my $left_side_coverage_item;
my $left_inside_coverage_item;

my @deletions;
my @is_list = @{$is_lists{$seq_id}};

for (my $i=1; $i<=$sequence_length; $i++)
{
	my @coverage_item = get_coverage(\$cov, $i);
	my $unique_coverage = $coverage_item[4]+$coverage_item[5];	
	my $repeat_coverage = $coverage_item[8]+$coverage_item[9];	

	#keep track of the current place where we dropped below the propagation cutoff
	if ($unique_coverage <= $deletion_propagation_cutoff)
	{	
		if (!defined $last_deletion_start_position)
		{
			$last_deletion_start_position = $i;
			$left_side_coverage_item = $last_coverage_item;
			$left_inside_coverage_item = \@coverage_item;
		}
	}
		
	##keep track of whether we've encountered the seed value
	if ($unique_coverage + $repeat_coverage <= $deletion_seed_cutoff)
	{
		$this_deletion_reaches_seed_value = 1;
	}

	##if we are above the propagation cutoff then record the current deletion
	if ( (defined $last_deletion_start_position) && ($unique_coverage > $deletion_propagation_cutoff) )
	{
		
		if ($this_deletion_reaches_seed_value)
		{
			my $last_deletion_end_position = $i-1;

			my $left_side_uniques = $left_side_coverage_item->[4]+$left_side_coverage_item->[5];
			my $left_inside_uniques = $left_inside_coverage_item->[4]+$left_inside_coverage_item->[5];

			my $right_side_uniques = $unique_coverage;
			my $right_inside_uniques = $last_coverage_item->[4]+$last_coverage_item->[5];
			
			my $size = $last_deletion_end_position - $last_deletion_start_position + 1;
			
			
			my $gene_string = '';
			my $feature_start = GetOverlappingFeature($last_deletion_start_position, \@is_list);
			$feature_start = GetOverlappingFeature($last_deletion_start_position, \@gene_list) if (!$feature_start);
			
			my $feature_end = GetOverlappingFeature($last_deletion_end_position, \@is_list);
			$feature_end = GetOverlappingFeature($last_deletion_end_position, \@gene_list) if (!$feature_end);
			
			if ($feature_start)
			{
				$gene_string .= "[$feature_start->{gene}]";
			}
			
			my $start_bound = ($feature_start) ? $feature_start->{end}-25 : $last_deletion_start_position;
			my $end_bound = ($feature_end) ? $feature_end->{start} : $last_deletion_end_position;

			foreach my $gene (@gene_list)
			{
				last if ($gene->{start} >= $end_bound);
				
				if ($gene->{start} > $start_bound)
				{
					$gene_string .= " " if ($gene_string);
					$gene_string .= "$gene->{gene}";
				}
			}
			
			## only add the end gene if it isn't the start gene
			if ($feature_end && (!$feature_start || ($feature_start != $feature_end)))
			{
				$gene_string .= " " if ($gene_string);
				$gene_string .= "[$feature_end->{gene}]";
			}
			
			##no genes found
			$gene_string .= "noncoding" if (!$gene_string);

			
			##Ok, take the gene names and concatenate everything between them... 
			
			my $line = "$last_deletion_start_position\t$last_deletion_end_position\t$size\t";
			$line .= "$left_side_uniques\t$left_inside_uniques\t$right_inside_uniques\t$right_side_uniques\t$gene_string";
			#print DEL "$line\n";
			
			my $del;
			$del->{start} = $last_deletion_start_position;
			$del->{end} = $last_deletion_end_position;
			$del->{line} = $line;
			
			push @deletions, $del;
		}
				
		#reset the search
		$this_deletion_reaches_seed_value = 0;
		undef $last_deletion_start_position;
	}
	
	$last_coverage_item = \@coverage_item;
}

##
# Gather together insertions and deletion SNPs that occur next to each other
##

my $lc;
for (my $i=0; $i<scalar @snps; $i++)
{
	my $c = $snps[$i];	 
	
	#if the same position and both are tandem insertion
	if ($lc && ($lc->{start} == $c->{start}) && ($lc->{insert_index}+1 == $c->{insert_index}))
	{
		$lc->{insert_index} = $c->{insert_index};
		$lc->{new_seq} .= $c->{new_seq};
		$lc->{quality} .= ",$c->{quality}";
		$lc->{total_coverage_string} .= ",$c->{total_coverage_string}";
		$lc->{best_coverage_string} .= ",$c->{best_coverage_string}";
		splice @snps, $i, 1;
		$i--;
		next;
	}
	
	#if the positions are next to each other and both are deletions...
	if ($lc && ($lc->{end}+1 == $c->{start}) && ($lc->{new_seq} eq '.') && ($c->{new_seq} eq '.') )
	{
		$lc->{end} = $c->{start};
		$lc->{ref_seq} .= $c->{ref_seq};
		$lc->{quality} .= ",$c->{quality}";
		$lc->{total_coverage_string} .= ",$c->{total_coverage_string}";
		$lc->{best_coverage_string} .= ",$c->{best_coverage_string}";
		splice @snps, $i, 1;
		$i--;
		next;
	}
	
	$lc = $c;
}

##
# Remove deletions that are in exactly the same positions as, or included within, SNPs
##

DEL: for (my $i=0; $i<scalar @deletions; $i++)
{
	my $del = $deletions[$i];	 
	
	SNP: foreach my $snp (@snps)
	{
		##remove the deletion if exactly the same as the SNP
		if ( ($del->{start} >= $snp->{start}) && ($del->{end} <= $snp->{end}) )
		{
			splice @deletions, $i, 1;
			$i--;
			last SNP;
		}
	}
}

##
# Write out deletion file
##

my $deletion_file_name = "deletions.tab";
open DEL, ">$deletion_file_name";
print DEL "start\tend\tsize\tleft\tleft_inside\tright_inside\tright\tgenes\n";
foreach my $del (@deletions)
{
	##remove the SNP if it is located within the deletion
	print DEL "$del->{line}\n";
}


##
# Remove SNPs that overlap deletions
##

SNP: for (my $i=0; $i<scalar @snps; $i++)
{
	my $snp = $snps[$i];
	DEL: foreach my $del (@deletions)
	{
		##remove the SNP if it is located within the deletion
		if ( ($del->{start} <= $snp->{start}) && ($del->{end} >= $snp->{end}) )
		{
			splice @snps, $i, 1;
			$i--;
			last DEL;
		}
	}
}




##
# Annotate SNPs
##

print STDERR "Annotating SNPs...\n";

#print Dumper @snps;
open ANN_SNPS, ">snps.annotated.tab";
print ANN_SNPS +join("\t", ("start", "end", "ref", "change", "quality", "cov", "tot_cov", "type", "mutation", "aa", "codon", "gene", "description"));
print ANN_SNPS "\n";

my $current_line;
SNP: foreach my $c (@snps)
{	
#	#correct the end coordinate
#	{
#		my $ref_rep_seq = ($c->{ref_seq} eq ".") ? "" : $c->{ref_seq};
#		#my $new_rep_seq = ($c->{new_seq} eq ".") ? "" : $c->{new_seq};
#		my $diff = length($ref_rep_seq) - length ($new_rep_seq);
#		$diff = 0 if ($diff < 0);
#		$c->{end} = $c->{start} + $diff;
#	}
	
	$current_line = join("\t", $c->{start}, $c->{end}, $c->{ref_seq}, $c->{new_seq}, $c->{quality}, $c->{best_coverage_string}, $c->{total_coverage_string}, '');
	
	##Determine whether the change is within a gene
	my $gene;
	my $prev_gene;
	my $next_gene;
	my $i;
	GENE: for ($i=0; $i < scalar @gene_list; $i++)
	{
		my $test_gene = $gene_list[$i];
		#print "$gene->{start} $gene->{end} $gene->{strand}\n";
		#This change is within a gene
		if ( ($test_gene->{start} <= $c->{start}) && ($test_gene->{end} >= $c->{end}) )
		{	
			$gene = $test_gene;
			last GENE;
		}
		#We've passed the changes, so it is in the previous intergenic space
		if ($test_gene->{start} > $c->{end})
		{
			$prev_gene = $gene_list[$i-1] if ($i > 0);
			$next_gene = $test_gene;
			last GENE;
		}
	}
	$prev_gene = $gene_list[-1] if (!$gene && !$prev_gene && !$next_gene);
	
	
	# Shift changes involving indels to the left (relative to top strand)
	# if the gene is on the opposite strand. Be aware that this could shift it outside of the gene.
	#
	# If it is an insertion, move if the inserted sequence is the same as ref sequence
	# if it is a deletion, move if deleted sequence is the same as ref seq
	# Problem, what if this change shifts it back to being noncoding?
	
	my $seq = $ref_seqs{$seq_id};
	if ($gene && ($gene->{strand} == - 1)) 
	{
		my $ref_seq = $c->{ref_seq};
		$ref_seq = '' if ($ref_seq eq '.');
		
		my $new_seq = $c->{new_seq};
		$new_seq = '' if ($new_seq eq '.');
		
		##deletion
		if (length($ref_seq) > length($new_seq))
		{
			my $len = length($ref_seq) - length($new_seq);
			if ($c->{start}-$len-1 >= 1)
			{
				my $test_shift_sequence = $seq->trunc($c->{start}-$len+1, $c->{start})->seq;
				MOVE: while ( "\U$test_shift_sequence" eq "\U$ref_seq")
				{
					if ($c->{start} - $len < 1)
					{
						last MOVE;
					}
					$c->{start} -= $len;
					$c->{end} -= $len;
					
					$test_shift_sequence = $seq->trunc($c->{start}-$len+1, $c->{start})->seq;
				}
			}
		}
		##insertion
		elsif (length($new_seq) > length($ref_seq))
		{
			my $len = length($new_seq) - length($ref_seq);
			if ($c->{start}-$len-1 >= 1)
			{
				my $test_shift_sequence = $seq->trunc($c->{start}-$len+1, $c->{start})->seq;
				#print STDERR +($c->{start}-$len-1) . " " . $c->{start} . "\n";
				#print STDERR "\U$test_shift_sequence" . " " . "\U$new_seq" . " " . "$c->{start}" . "\n";
				MOVE: while ("\U$test_shift_sequence" eq "\U$new_seq")
				{
					if ($c->{start} - $len < 1)
					{
						last MOVE;
					}
					$c->{start} -= $len;
					$c->{end} -= $len;

					$test_shift_sequence = $seq->trunc($c->{start}-$len+1, $c->{start})->seq;
					
					#print STDERR +($c->{start}-$len-1) . " " . $c->{start} . "\n";
					#print STDERR "\U$test_shift_sequence" . " " . "\U$new_seq" . " " . "$c->{start}" . "\n";
				}
			}
		}
		# There is a chance that the change is now noncoding!!!!
		if ($c->{end} < $gene->{start})
		{
			$gene = undef;
			$next_gene = $gene_list[$i];
			$prev_gene = $gene_list[$i-1] if ($i > 0);
		}
		
		
		##we have to rewrite the initial part of the line
		$current_line = join("\t", $c->{start}, $c->{end}, $c->{ref_seq}, $c->{new_seq}, $c->{quality}, $c->{best_coverage_string}, $c->{total_coverage_string}, '');

	}	
	
	
	## Change is noncoding, $prev_gene and/or $next_gene will be defined
	if (!$gene)
	{	
		$current_line .= "noncoding\t\t\t\t";
		if ($prev_gene)
		{
			$current_line .= "$prev_gene->{gene} ";
			$current_line .= +(($prev_gene->{strand} == +1) ? ">" : "<");
			$current_line .= " " . ($c->{start} - $prev_gene->{end});
		}
		$current_line .= " // ";
		if ($next_gene)
		{
			$current_line .= +($next_gene->{start} - $c->{end}) . " ";
			$current_line .= +(($next_gene->{strand} == +1) ? ">" : "<");
			$current_line .= " $next_gene->{gene}";
		}
		$current_line .=  "\t";
		$current_line .=$prev_gene->{product} if ($prev_gene); 
		$current_line .= "//"; 
		$current_line .= $next_gene->{product} if ($next_gene);

		next SNP;
	}
	
	
	## Change is coding

	
	##Change is in a non-protein gene
	##or a pseudogene
	##index by nucleotide number
	if (!$gene->{translation})
	{
		my $nt_num;
		my $ref_nt = $c->{ref_seq};
		my $new_nt = $c->{new_seq};		
		
		if ($gene->{strand} == +1)
		{
			$nt_num = $c->{start} - $gene->{start} + 1;
		}
		else
		{
			$nt_num = $gene->{end} - $c->{end} + 1;
			
			#remember to revcom the changes
			if ($c->{ref_seq} ne '.')
			{
				my $t = new Bio::Seq(-seq => $c->{ref_seq});
				$c->{ref_seq} = $t->revcom()->seq();
			}
			if ($c->{new_seq} ne '.')
			{
				my $t = new Bio::Seq(-seq => $c->{new_seq});
				$c->{new_seq} = $t->revcom()->seq();
			}
		}		
		$current_line .= "$gene->{type}\t\t\t";
		$current_line .= "$ref_nt" . "$nt_num" . "$new_nt\t";
		$current_line .= "$gene->{gene}\t$gene->{product}";		
		
		next SNP;
	}
		
	#Single base substitution - determine if (non)synonymous.
	if (!($c->{ref_seq} =~ m/\./) && !($c->{new_seq} =~ m/\./) && (length $c->{ref_seq} == 1) && (length $c->{new_seq} == 1))
	{
		my $codon_num = 0; #Numbering starting at zero
		my $codon_pos = 0;
		my $codon_seq;
		my $ref_codon;
		my $new_codon;
		my $ref_aa;
	    my $new_aa;
		if ($gene->{strand} == +1)
		{
			$codon_num = int(($c->{start}-$gene->{start})/3) + 1;
			$codon_pos = ($c->{start}-$gene->{start}) % 3;
			$codon_seq = $seq->trunc($gene->{start} + 3 * ($codon_num-1),$gene->{start} + 3 * $codon_num - 1);
			
			$ref_codon = $codon_seq->seq();
			$ref_aa = $codon_seq->translate( undef, undef, undef, 11 )->seq();
			
			$new_codon = $codon_seq->seq();
			substr($new_codon, $codon_pos, 1) = $c->{new_seq};
			$codon_seq->seq($new_codon);
			$new_aa = $codon_seq->translate( undef, undef, undef, 11 )->seq();
		}
		else #reverse strand
		{
			$codon_num = int(($gene->{end}-$c->{start})/3) + 1;
			$codon_pos = ($gene->{end}-$c->{start}) % 3;
			$codon_seq = $seq->trunc($gene->{end} - 3 * $codon_num+1,$gene->{end} - 3 * ($codon_num-1));
			$codon_seq = $codon_seq->revcom();
			
			#remember to revcom the changes
			my $t = new Bio::Seq(-seq => $c->{ref_seq});
			$c->{ref_seq} = $t->revcom()->seq();
			$t = new Bio::Seq(-seq => $c->{new_seq});
			$c->{new_seq} = $t->revcom()->seq();
			
			$ref_codon = $codon_seq->seq();
			$ref_aa = $codon_seq->translate( undef, undef, undef, 11 )->seq();
			
			$new_codon = $codon_seq->seq();
			substr($new_codon, $codon_pos, 1) = $c->{new_seq};
			$codon_seq->seq($new_codon);
			$new_aa = $codon_seq->translate( undef, undef, undef, 11 )->seq();
		}
		
		$current_line .= "$gene->{type}\t" . (($ref_aa eq $new_aa) ? "S" : "NS") . "\t";
		$current_line .= "$ref_aa" . $codon_num . "$new_aa\t$ref_codon => $new_codon\t";
		$current_line .= "$gene->{gene}\t$gene->{product}";
		
		next SNP;
	}
	
	
	#Deletions, insertions, and frameshifts
	#Codes are INS (in-frame insertion), DEL (in-frame deletion), FS (frameshift)
	{
		my $gene_start = $gene->{start};
		my $gene_end = $gene->{end};
		
		my $ref_seq = $c->{ref_seq};
		my $new_seq = $c->{new_seq};

		#Make seqs the actual things we want to use in replacements
		my $ref_rep_seq = ($c->{ref_seq} eq ".") ? "" : $c->{ref_seq};
		my $new_rep_seq = ($c->{new_seq} eq ".") ? "" : $c->{new_seq};
		
		## Change does not result in frameshift
		if (abs(length($new_rep_seq) - length($ref_rep_seq)) % 3 == 0)
		{
			## Insertion of amino acids
			my ($codon_num, $ref_codon_seq, $new_codon_seq);
			$ref_codon_seq =  Bio::Seq->new( '-seq' => $ref_rep_seq) if (length $ref_rep_seq > 0);
			$new_codon_seq =  Bio::Seq->new( '-seq' => $new_rep_seq) if (length $new_rep_seq > 0);
						
			if ($gene->{strand} == +1)
			{
				$codon_num = int(($c->{start}-$gene->{start})/3) + 1;
			}
			else
			{
				$codon_num = int(($gene->{end}-$c->{start})/3) + 1;
				$ref_codon_seq = $ref_codon_seq->revcom if ($ref_codon_seq);
				$new_codon_seq = $new_codon_seq->revcom if ($new_codon_seq);
			}
			
			my ($ref_codon, $new_codon) = ('.','.');
			$ref_codon = $ref_codon_seq->seq if ($ref_codon_seq);
			$new_codon = $new_codon_seq->seq if ($new_codon_seq);

			## translate to the inserted or deleted amino acid sequence
			my ($ref_aa, $new_aa) = ('.','.');
			$ref_aa = $ref_codon_seq->translate( undef, undef, undef, 11 )->seq if ($ref_codon_seq);
			$new_aa = $new_codon_seq->translate( undef, undef, undef, 11 )->seq if ($new_codon_seq);
			
			## create the output line
			$current_line .= "$gene->{type}\t" . (length $new_rep_seq > length $ref_rep_seq) ? "INS" : "DEL";
			$current_line .= "\t$ref_aa" . $codon_num . "$new_aa\t$ref_codon => $new_codon\t";
			$current_line .= "$gene->{gene}\t$gene->{product}";
			
		} 
		## Change results in a frameshift
		else 
		{
			my ($codon_num, $ref_reading_frame, $new_reading_frame, $ref_orf_seq, $new_orf_seq);
							
			if ($gene->{strand} == +1)
			{
				$codon_num = int(($c->{start}-$gene->{start})/3) + 1;
				
				$ref_reading_frame = $seq->trunc($gene->{start}, $gene->{end});
				$ref_orf_seq = $ref_reading_frame->seq();
				$new_orf_seq = substr($ref_orf_seq, 0, $c->{start}-$gene->{start}+1) . $new_rep_seq 
					. substr($ref_orf_seq, $c->{start}+length($ref_rep_seq)-$gene->{start}+1, $gene->{end}-($c->{start}+length($ref_rep_seq)));
		
				#extend the reading frame so we can find the new protein length (and eventually sequence)
				if ($gene->{end}+1 <= $seq->length)
				{
					my $safe_end = $gene->{end}+10000;
					$safe_end = $seq->length if ($safe_end > $seq->length);
					$new_orf_seq .=  $seq->trunc($gene->{end}+1, $safe_end)->seq;
				}
							
				##ok, finally have the new reading frame
				$new_reading_frame = Bio::Seq->new( '-seq' => $new_orf_seq);
			}
			else #Reverse strand
			###
			### Add code to shift differences over so they show up as early in the reading frame as possible!!!
			### Be sure that this works for insertions with length > 1 as well.
			###
			
			{
				$codon_num = int(($gene->{end}-$c->{start})/3) + 1;
				
				$ref_reading_frame = $seq->trunc($gene->{start}, $gene->{end});
				$ref_orf_seq = $ref_reading_frame->seq();
				$new_orf_seq = substr($ref_orf_seq, 0, $c->{start}-$gene->{start}+1) . $new_rep_seq
					. substr($ref_orf_seq, $c->{start}+length($ref_rep_seq)-$gene->{start}+1, $gene->{end}-($c->{start}+length($ref_rep_seq)));

				#extend the reading frame so we can find the new protein length (and eventually sequence)
				if ($gene->{start}-1 >= 1)
				{
					my $safe_start = $gene->{start}-10000;
					$safe_start = 1 if ($safe_start <= 1);
					$new_orf_seq =  $seq->trunc($safe_start, $gene->{start}-1)->seq . $new_orf_seq;
				}
				$new_reading_frame = Bio::Seq->new( '-seq' => $new_orf_seq);

				$ref_reading_frame = $ref_reading_frame->revcom();
				$new_reading_frame = $new_reading_frame->revcom();
			}		
			
			## determine how the length of the ORF changes
			my $ref_protein = $ref_reading_frame->translate( undef, undef, undef, 11 )->seq;
		#	die "Did not find a stop codon in reference protein" if (! ($ref_protein =~ s/\*.+/\*/) );		
			my $ref_length = length($ref_protein) - 1;

			my $new_protein = $new_reading_frame->translate( undef, undef, undef, 11 )->seq;
			die "Did not find a stop codon in new protein" if (! ($new_protein =~ s/\*.+/\*/) );		
			my $new_length = length($new_protein) - 1;
			
			
			#print STDERR Dumper($ref_reading_frame);
			#print STDERR Dumper($new_reading_frame);
			#print STDERR $ref_protein . "\n";
			#print STDERR $new_protein . "\n";

			my $ref_aa = substr $ref_protein, $codon_num-1, 1;
			my $new_aa = substr $new_protein, $codon_num-1, 1;

			$current_line .= "$gene->{type}\tFS\t$ref_aa$codon_num$new_aa\t$ref_length" . " aa => " . $new_length . " aa\t$gene->{gene}\t$gene->{product}";
			$current_line .= "\t$ref_protein\t$new_protein";
			#$current_line .= "\t" . $ref_reading_frame->seq();
			#$current_line .= "\t" . $new_reading_frame->seq();
			#$current_line .= "\t";
			#$current_line .= "$gene->{gene}\t$gene->{product}";
		}
		next SNP;	
	}	
} continue {
	print ANN_SNPS "$current_line\n";
}

##
# Print Deletions
##


##
# Utility Functions
##


sub quality_to_char
{
	my ($q) = @_;
	return chr($q + 33);
}

sub char_to_quality
{
	my ($c) = @_;
	return (unpack("C",$c) - 33);
}


## calls the correct function for the mode we are using
## also corrects for the different offset used by Solexa of 64+qual vs usual 33+qual for Phred-style
sub quality_to_error_probability
{
	my ($q) = @_;
	return ($using_solexa_quality) ? (solexa_quality_to_error_probability($q-31)) : (phred_quality_to_error_probability($q));
}

sub phred_quality_to_error_probability
{
	my ($q) = @_;
	my $val = exp (-$q * log(10) / 10);
	return $val;
}

sub solexa_quality_to_error_probability
{
	my ($q) = @_;
	my $val = exp($q/10 * log(10));
	return 1 / (1 + $val);
}

sub GetTag
{
	my ($Feature, $Tag, $Allow_Array) = @_;
	$Allow_Array = 0 if (!defined $Allow_Array);
	
	return '' if (!$Feature->has_tag("$Tag"));
	my @Tag_Values = $Feature->get_tag_values("$Tag");
	return '' if (scalar @Tag_Values == 0);
	return $Tag_Values[0] if (!$Allow_Array);
	return @Tag_Values;
}	

sub GetOverlappingFeature
{
	my ($pos, $feature_list_ref) = @_;
	foreach my $f (@$feature_list_ref)
	{
		return $f if ( ($pos >= $f->{start}) && ($pos <= $f->{end}) );
	}
	return undef;
}

sub pack_snp
{
	my ($position, $insert_count, $quality, $strand, $redundancy, $base) = @_;
	#print "($quality, $strand, $redundancy, $pair, $base) \n";
	my $converted_strand = ($strand == +1) ? 1 : 0;
	my $converted_base = $base_to_int{$base};
	
	#treat any other character as an 'N'.
	$converted_base = $base_to_int{'N'} if (!defined $converted_base);
	
	die "Redundancy overflow: $redundancy" if ($redundancy > unpack("S", 0xFFFF));
	return pack("LCCCS", $position, $insert_count, $converted_strand + $converted_base*2, $quality, $redundancy);
}

sub unpack_snp
{
	my ($snp) = @_;
	my ($position, $insert_count, $coded, $quality, $redundancy) = unpack("LCCCS", $snp);
		
	my $strand = ($coded % 2 == 1) ? +1 : -1;
	$coded = int($coded/2);
	my $base = $int_to_base[$coded];
	
	return ($position, $insert_count, $quality, $strand, $redundancy, $base);
}

sub get_snp_diff
{
	my ($snp_ref, $pos) = @_;
	return unpack_snp( substr($$snp_ref, $pos*$snp_diff_item_size, $snp_diff_item_size) );
}


sub add_coverage
{
	my ($pos, $add) = @_;
	
	##read the old entry
	#print "$cov, $add, $coverage_item_format, $coverage_item_size \n";
	
	my @item = unpack $coverage_item_format, (substr $cov, ($pos-1)*$coverage_item_size, $coverage_item_size);
	
	#print STDERR "@item\n";
	
	for (my $i=0; $i<scalar @item; $i++)
	{
		next if (!defined $add->[$i]);
	
		## the first values are currently quality scores, need to convert to probabilities
		$item[$i] += $add->[$i];
	}
	
	#print STDERR "@item\n";
	substr($cov, ($pos-1)*$coverage_item_size, $coverage_item_size) = pack($coverage_item_format, @item);
}

sub get_coverage
{
	my ($cov_ref, $pos) = @_;
	return unpack $coverage_item_format, substr($$cov_ref, ($pos-1)*$coverage_item_size, $coverage_item_size);
}


sub filter_homopolymeric
{
	## read // reference to sequence which has been split into a list // length below which to do filtering
	my ($match_group_ref) = @_;
	my $length_cutoff = 22;
	
	my @filtered_matches;
	MATCH_GROUP: foreach my $m (@$match_group_ref)
	{
		##don't examine matches that are above a certain length threshold to save time 
		if ($m->{matches}->[0]->{reference_end} - $m->{matches}->[0]->{reference_start} + 1 <= $length_cutoff)
		{
			#my $subseq = $seq->subseq($m->{reference_start}, $m->{reference_end});
			my @subseq = split //, $m->{ref_seq_matched}->{1};
			
			my %nt_hash;
			foreach my $nt (@subseq)
			{
				$nt_hash{$nt}++;
			}
			my @keys = sort { $nt_hash{$a} <=> $nt_hash{$b} } keys %nt_hash;
			#my $len = length $subseq;
			my $len = scalar @subseq;
			my $top = $nt_hash{$keys[-1]};
			
			my $diff_nt = scalar @keys;
			if ($diff_nt <= 2)
			{
				#print STDERR "Only two nucleotides: $m->{reference_start}\-$m->{reference_end}  $diff_nt\n";
				#print STDERR "@subseq\n";
				next MATCH_GROUP;				
			}
			
			if ($top / $len > 0.8) 
			{
				#print STDERR "Homopolymer: $m->{reference_start}\-$m->{reference_end}  $top / $len\n";
				#print STDERR "@subseq\n";
				next MATCH_GROUP;
			}	
		}
				
		push @filtered_matches, $m;
	}
	
	return @filtered_matches;
}

